/********************************************************************************
 *   Jabberoo/Judo -- C++ Jabber Library                                        *
 *                                                                              * 
 *   Copyright (C) 1999-2000 Dave Smith (dave@jabber.org)                       *
 *                                                                              *
 *   This library is free software; you can redistribute it and/or              *
 *   modify it under the terms of the GNU Lesser General Public                 *
 *   License as published by the Free Software Foundation; either               *
 *   version 2.1 of the License, or (at your option) any later version.         *
 *                                                                              *
 *   This library is distributed in the hope that it will be useful,            *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
 *   Lesser General Public License for more details.                            *
 *                                                                              *
 *   You should have received a copy of the GNU Lesser General Public           *
 *   License along with this library; if not, write to the Free Software        *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
 ********************************************************************************
 */
#include "jabberoo.hh"
#include "jabberoox.hh"
using namespace jabberoo;

// ---------------------------------------------------------
//
// Agent methods
//
// ---------------------------------------------------------
Agent::Agent(const Agent& a)
     : _session(a._session), _jid(a._jid), _name(a._name),
       _desc(a._desc), _service(a._service), _transport(a._transport),
       _registerable(a._registerable), _searchable(a._searchable),
       _subagents(a._subagents)
{
     if (_name.empty())
	  _name = "No name available (" + a._jid + ")";
}

Agent::Agent(const string& jid, const Tag& baseTag, Session& s)
     : _jid(jid), _session(s)
{
     _name      = baseTag.getTaggedCDATA("name");
     _desc      = baseTag.getTaggedCDATA("description");
     _service   = baseTag.getTaggedCDATA("service");
     _transport = baseTag.getTaggedCDATA("transport");
     _registerable   = (baseTag.getTag("register") != NULL);
     _searchable     = (baseTag.getTag("search") != NULL);
     _subagents      = (baseTag.getTag("agents") != NULL);
     _gc_capable     = (baseTag.getTag("groupchat") != NULL);
     _fetchrequested = false;
}

Agent::~Agent()
{}

inline const string& Agent::JID() const
{
     return _jid;
}

inline const string& Agent::name() const
{
     return _name;
}

inline const string& Agent::description() const
{
     return _desc;
}

inline const string& Agent::service() const
{
     return _service;
}

inline const string& Agent::transport() const
{
     return _transport;
}

inline bool Agent::isRegisterable() const
{
     return _registerable;
}

inline bool Agent::isSearchable() const
{
     return _searchable;
}

inline bool Agent::isGCCapable() const
{
     return _gc_capable;
}

inline bool Agent::hasAgents() const
{
     return _subagents;
}

void Agent::requestRegister(TagCallbackFunc f)
{
     _session.queryNamespace("jabber:iq:register", f, _jid);
}

void Agent::requestSearch(TagCallbackFunc f)
{
     _session.queryNamespace("jabber:iq:search", f, _jid);
}

void Agent::fetch()
{
     if ( _subagents && empty() && !_fetchrequested )
     {
	  // Release existing agents
	  clear();
	  // Set the fetch request flag
	  _fetchrequested = true;
	  // Request agents
	  _session.queryNamespace("jabber:iq:agents", slot(this, &Agent::on_fetch), _jid);
     }
}         

void Agent::on_fetch(const Tag& iq)
{
     // Reset fetch request flag
     _fetchrequested = false;

     // IQ should contain an <iq> tag with one child tag <query>; within <query>,
     // there should be 1 or more <agent> tags -- these tags must be parsed into
     // agent objects and stored internally to "this" agent
     Tag* query = iq.getTag("query");
     if (query == NULL)
	  return;

     if (iq.cmpAttrib("type", "result"))
     {
	  // Walk the <query> tag's children
	  ElementList::const_iterator it = query->getChildren().begin();
	  for (; it != query->getChildren().end(); it++)
	  {
	       // Cast the child element into a tag
	       Tag& t = *static_cast<Tag*>(*it);
	       
	       // Ensure this is an agent tag
	       if (t.getName() == "agent")
		    push_back(Agent(t.getAttrib("jid"), t, _session));
	  }

	  // Notify any one-time observers that this agent has been updated successfully
	  evtFetchComplete(true);
     }
     else
     {
	  // Notify any observer that this agent has not been update successfully
	  evtFetchComplete(false);
     }

     // Release all one-time observers
     evtFetchComplete.clear();
}

AgentList::AgentList(const Tag& query, Session& s)
{
     load(query, s);
}

void AgentList::load(const Tag& query, Session& s)
{
     // Ensure list is clear
     clear();
     // MAke sure this is the proper namespace
     if (!query.cmpAttrib("xmlns", "jabber:iq:agents"))
	  return;

     // Walk each of the child <agent> tags and build agent objects
     ElementList::const_iterator it = query.getChildren().begin();
     for (; it != query.getChildren().end(); it++)
     {
	  // Cast the child element to a tag
	  Tag& t = *static_cast<Tag*>(*it);

	  // Ensure this is an <agent> tag
	  if (t.getName() == "agent")
	       // Construct and insert an agent
	       push_back(Agent(t.getAttrib("jid"), t, s));
     }
}
